local g_ments = {}
local g_delays = {}
local g_mentColors = {}

-- cocos로 부터 전달 받은 값
local g_homeTeam = {}
local g_awayTeam = {}

local g_homePlayers = {}
local g_awayPlayers = {}

local g_homePositions = {}
local g_awayPositions = {}

local g_homePitcherId
local g_awayPitcherId

local g_homeCpId
local g_awayCpId

local g_isHome

local g_replayData
-------------------------------------
local g_inning
local g_isUp

local g_currBatter
local g_currPitcher
local g_currBatterId

local g_lineIter
local g_commands

local g_mentQueue

local g_homeScore
local g_awayScore

local g_ballCnt
local g_strikeCnt
local g_outCnt

local g_bases

local g_state
local g_hitContext

local g_history

local g_fgColor
local g_bgColor
local g_mentColor

local g_pickoff

-- 3아웃일때 주자가 다음베이스로 진루하는것을 막기 위한 플래그. 병살일 때 2루주자가 홈까지 들어오는게 어색해서 사용됨
local g_preventForceRun

local STATE_INIT        = 0
local STATE_SKIP        = 1
local STATE_HIT         = 2
local STATE_SBNT        = 3
local STATE_STL         = 4
local STATE_BAT_AGAIN   = 5
local STATE_PITWP       = 6
local STATE_NOTOUT      = 7
local STATE_SO          = 8
local STATE_BB          = 9
local STATE_HBP         = 10
local STATE_GAMEOVER    = 11
local STATE_PICKOFF     = 12

local BLINK = "_BLINK"

local g_toInning

local g_stealer
local g_decideRun

local g_isPause
local g_waitResume
local g_isStop

local g_posName = {
    P = "투수",
    C = "포수",
    B1 = "1루수",
    B2 = "2루수",
    B3 = "3루수",
    SS = "유격수",
    LF = "좌익수",
    CF = "중견수",
    RF = "우익수"
}

local g_baseName = {
    "1루", "2루", "3루", "홈"
}

local g_fenceName = {
    LF = "좌측 담장",
    CF = "가운데 담장",
    RF = "우측 담장"
}

function clearObject(t)
    for k in pairs(t) do
        rawset(t, k, nil)
    end
end

function copyObject(src, dst)
    for k, v in pairs(src) do
        rawset(dst, k, v)
    end
end

function dumpObject(t)
    for k, v in pairs(t) do
        print(k .. " => " .. v)
    end
end

function clearPlayers(t)
    for k in pairs(t) do
        rawset(t, k, nil)
    end
end

function copyPlayers(src, dst)
    for k, v in pairs(src) do
        rawset(dst, k, v)
    end
end

function dumpPlayers(t)
    for k, v in pairs(t) do
        print(k .. " => " .. v.name .. " " .. v.playerType)
    end
end

function getDefPlayer(pos)
    if pos == "P" then
        return g_currPitcher
    end

    if g_isUp then
        return g_homePlayers[g_homePositions[pos]]
    else
        return g_awayPlayers[g_awayPositions[pos]]
    end
end

function init(homeTeam, awayTeam, homePlayers, awayPlayers, homePositions, awayPositions, homePitcherId, awayPitcherId, homeCpId, awayCpId, isHome, replyData, seed)
    clearObject(g_homeTeam)
    clearObject(g_awayTeam)
    
    copyObject(homeTeam, g_homeTeam)
    copyObject(awayTeam, g_awayTeam)

    -- dumpObject(g_homeTeam)
    -- dumpObject(g_awayTeam)
    
    clearObject(g_homePositions)
    clearObject(g_awayPositions)

    copyObject(homePositions, g_homePositions)
    copyObject(awayPositions, g_awayPositions)
    
    clearPlayers(g_homePlayers)
    clearPlayers(g_awayPlayers)

    copyPlayers(homePlayers, g_homePlayers)
    copyPlayers(awayPlayers, g_awayPlayers)

    -- dumpPlayers(g_homePlayers)
    -- dumpPlayers(g_awayPlayers)
    
    g_homePitcherId = homePitcherId
    g_awayPitcherId = awayPitcherId
    
    g_homeCpId = homeCpId
    g_awayCpId = awayCpId
    
    g_isHome = isHome

    g_replayData = replyData
    
    -- 초기화
    g_inning = 1;
    g_isUp = true;

    g_homeScore = 0
    g_awayScore = 0
    
    g_ballCnt = 0
    g_strikeCnt = 0
    g_outCnt = 0

    g_commands = {}
    g_mentQueue = {}
    
    g_bases = {}
    g_bases[1] = {finished = true}
    g_bases[2] = {finished = true}
    g_bases[3] = {finished = true}
    g_bases[4] = {finished = true}
    
    g_state = STATE_INIT

    g_decideRun = false
    
    g_isPause = false
    g_waitResume = false
    g_isStop = false
    
    g_history = nil

    g_preventForceRun = false

    g_fgColor = g_awayTeam.fgColor;
    g_bgColor = g_awayTeam.bgColor;
    
    -- 리플레이 데이터를 한줄 단위로 가져오는 iteration function
    g_lineIter = string.gmatch(g_replayData, "([^\n]*)%s*")

    -- 동일한 중계멘트를 보여주기 위한 랜덤시드 설정
    math.randomseed(seed)

    collectgarbage("collect")

    --g_state = STATE_SKIP
    --g_toInning = {inning = 1, dir = "D"}
end

function setPause(isPause)
    g_isPause = isPause
    
    if not g_isPause and g_waitResume then
        g_waitResume = false
        
        proceed()
    end
end

function stop()
    -- 이미 중계가 종료되었을 때는 stop플래그를 설정하지 않음. 그렇지 않을 경우 다시 skipInning이 호출되어 문제가 발생
    if g_state == STATE_GAMEOVER then
        return
    end
    
    g_isStop = true
    
    if g_isPause and g_waitResume then
        g_waitResume = false
        
        proceed()
    end
end

-- 한 line에서 command 리스트를 파싱해서 반환
function getCommands(line)
    local commands = {}
    
    local pattern = "([^\n (]+)%(([^)]*)%)"
    for code, params in string.gmatch(line, pattern) do
        local command = {}
        command.code = code
        command.params = {}

        if #params > 0 then
            for param in string.gmatch(params, "[^,]+") do
                table.insert(command.params, param)
            end
        end

        table.insert(commands, command)
    end

    return commands
end

function removeCommands()
    for i, v in ipairs(g_commands) do
        g_commands[i] = nil
    end
end

function skipInnings(inning, isUp)
    while true do
        if #g_commands == 0 then
            local line = g_lineIter()
            g_commands = getCommands(line)
        end
        
        local c = table.remove(g_commands, 1)
        if c.code == "GAMEOVER" then
            live_call("setGameover", g_inning, g_isUp)
            return
        elseif c.code == "INN" then
            g_inning = tonumber(c.params[1])
            g_isUp = c.params[2] == "U"
            
            -- 목적 이닝에 도달했으면 return
            if g_inning == inning and isUp == g_isUp then
                g_state = STATE_INIT
                processInning(c)
                return
            end
        elseif (c.code == "RUN" and c.params[2] == "4") or (c.code == "BRUN" and c.params[1] == "4") then
            if c.code == "RUN" and c.params[3] == "W" then
                -- 아직 득점이 난 상황이 아님. 1루 기다림
                g_decideRun = true
            else
                if g_isUp then
                    g_awayScore = g_awayScore + 1
                else
                    g_homeScore = g_homeScore + 1
                end
            end
        elseif c.code == "SCORE" then
            live_call("setScore", g_inning, g_isUp, tonumber(c.params[3]))
        elseif c.code == "OUT" then
            g_outCnt = tonumber(c.params[1])
        elseif c.code == "HT" then
            if g_decideRun then
                if g_outCnt < 3 then
                    -- 득점 인정
                    if g_isUp then
                        g_awayScore = g_awayScore + 1
                    else
                        g_homeScore = g_homeScore + 1
                    end
                end

                g_decideRun = false
            end
        elseif c.code == "H" or c.code == "E" or c.code == "B" then
            updateScoreboard(c)
        elseif c.code == "PITCNG" then
            -- 투수 교체
            if g_isUp then
                g_homePitcherId = c.params[1]
            else
                g_awayPitcherId = c.params[1]
            end

            live_call("setPitcher", g_isUp, tonumber(c.params[1]))
        end
    end
end

-- 메인 루프
function proceed()
    --print("[proceed]")

    if g_isStop then
        skipInnings(0, false)
        --print("stop finished " .. g_inning .. " " .. tostring(g_isUp))
        return
    elseif g_isPause then
        g_waitResume = true
        return
    end

    if g_state == STATE_SKIP then
        skipInnings(g_toInning.inning, g_toInning.dir == "U")
        return
    elseif g_state == STATE_HIT then
        postProcessHit()
        return
    elseif g_state == STATE_SBNT then
        postProcessSbnt()
        return
    elseif g_state == STATE_STL then
        postProcessStl()
        return
    elseif g_state == STATE_PICKOFF then
        postProcessPickoff()
        return
    elseif g_state == STATE_PITWP then
        postProcessPitwp()
        return
    elseif g_state == STATE_NOTOUT then
        postProcessNotout()
        return
    elseif g_state == STATE_SO then
        postProcessSo()
        return
    elseif g_state == STATE_BB then
        postProcessBb()
        return
    elseif g_state == STATE_HBP then
        postProcessHbp()
        return
    end

    -- 남은 멘트가 있으면 멘트 보여주기
    while #g_mentQueue > 0 do
        local obj = table.remove(g_mentQueue, 1)
        
        -- 공백 멘트일 경우 다음 멘트 보여주기
        if showMent(obj.mentCode, obj.params, obj.options) then
            return
        end
    end

    if g_state == STATE_GAMEOVER then
        live_call("setGameover", g_inning, g_isUp)
        return
    end

    -- 처리해야할 커맨드가 없을 경우 다음 라인에서 커맨드를 얻어옴
    if #g_commands == 0 then
        local line = g_lineIter()
        g_commands = getCommands(line)
    end
    
    while #g_commands > 0 do
        local c = table.remove(g_commands, 1)
        --print("command : " .. c.code)
        
        if c.code == "START" then
            processStart()
        elseif c.code == "INN" then
            processInning(c)
        elseif c.code == "BAT" then
            processBat(c)
            --break
        elseif c.code == "C" then
            if processBallCount(c) then
                break
            end
        elseif c.code == "HIT" then
            processHit(c)
            
            -- 에러일 경우 P가 없기 때문에 강제로 끊어줌
            if g_hitContext.isError then
                break
            end
        elseif c.code == "SO" then
            processSo(c)
        elseif c.code == "BB" then
            processBb(c)
            -- P가 없기 때문에 강제로 끊어줌
            break
        elseif c.code == "HBP" then
            processHbp(c)
            -- P가 없기 때문에 강제로 끊어줌
            break
        elseif c.code == "SBNT" or c.code == "SBNT-" then
            processSbnt(c)
        elseif c.code == "PITCNG" then
            processPitcng(c)
        elseif c.code == "STL" or c.code == "STL-" then
            processStl(c)
            break
        elseif c.code == "POFF" then
            processPickoff(c)
        elseif c.code == "PITWP" or c.code == "CATPB" then
            processPitwp(c)
        elseif c.code == "NOTOUT" or c.code == "NOTOUT-" then
            processNotout(c)
        elseif c.code == "HT" then
            --removeCommands()
            -- 다음 라인 불러옴
            --local line = g_lineIter()
            --g_commands = getCommands(line)
        elseif c.code == "P" then
            break
        elseif c.code == "GAMEOVER" then
            processGameover()
            break
      	elseif c.code == "BS" or c.code == "BT" or c.code == "PS" or c.code == "PT" then
            -- 투수 교체로 인하여 타자/투수 스킬, 전술이 다시 나올 수 있음
            -- processSkills 메소드에서 처리를 위하여 commands에 다시 추가
            table.insert(g_commands, 1, c)
            
            local batterSkills, pitcherSkills, records = processSkills()
            
            -- 타자, 투수 스킬 보여주기
            showBatterSkills(batterSkills)
            showPitcherSkills(pitcherSkills)
            
            -- 다음 P 강제 삭제. 그렇지 않을 경우 중계 멈춤
            if #g_commands > 0 then
              local nextCommand = g_commands[1]
              if nextCommand.code == "P" then
                table.remove(g_commands, 1)
              end
            end
            
            --[[
            while #g_commands > 0 do
                c = table.remove(g_commands, 1)
                if c.code == "P" then
                    break
                end
            end
            --]]
        else
            -- 에러 발생
            return
            --error("unknown command : " .. c.code)
        end
    end
end

-- 게임 시작
function processStart()
    local myTeamName = g_isHome and g_homeTeam.teamName or g_awayTeam.teamName
    local otherTeamName = g_isHome and g_awayTeam.teamName or g_homeTeam.teamName

    showMent("START", {myTeamName = myTeamName, otherTeamName = otherTeamName})
end

-- 이닝
function processInning(command)
    g_inning = tonumber(command.params[1])
    g_isUp = command.params[2] == "U"
    
    -- 현재 투수 세팅
    g_currPitcher = g_isUp and g_homePlayers[g_homePitcherId] or g_awayPlayers[g_awayPitcherId]

    local teamName = g_isUp and g_awayTeam.teamName or g_homeTeam.teamName
    
    -- 이닝 해설 멘트 추가
    if isMyAtk() then
        -- 점수 차 계산
        local diff
        if g_isHome then
            diff = g_homeScore - g_awayScore
        else
            diff = g_awayScore - g_homeScore
        end

        if diff >= 4 then
            addMentQueue("INN_1")
        elseif diff >= 1 and diff >= 3 then
            addMentQueue("INN_2")
        elseif diff == 0 then
            if g_homeScore == 0 then
                addMentQueue("INN_3")
            else
                addMentQueue("INN_4")
            end
        elseif diff <= -1 and diff >= -3 then
            addMentQueue("INN_5")
        elseif diff <= -4 then
            addMentQueue("INN_6")
        end
    end

    -- 이닝 알림
    live_call("showInning", g_inning, g_isUp, "INNING")
    
    showMent("INN", {inning = g_inning, dir = g_isUp and "초" or "말", teamName = teamName})
    
    -- 주자 초기화
    g_bases[1] = {finished = true}
    g_bases[2] = {finished = true}
    g_bases[3] = {finished = true}
    g_bases[4] = {finished = true}
    
    g_history = nil

    g_preventForceRun = false

    removeCommands()
end

function processSkills()
    local batterSkills = {}
    local pitcherSkills = {}
    local records = {}

    while #g_commands > 0 do
        local nextCommand = g_commands[1]
        if nextCommand.code == "BS" or nextCommand.code == "BT" then
            table.remove(g_commands, 1)
            
            -- 타자 스킬, 작전에 넣어줌
            for i, v in ipairs(nextCommand.params) do
                v = string.gsub(v, "_", " ")
                table.insert(batterSkills, v)
            end
        elseif nextCommand.code == "PS" or nextCommand.code == "PT" then
            table.remove(g_commands, 1)
            
            -- 투수 스킬, 작전에 넣어줌
            for i, v in ipairs(nextCommand.params) do
                v = string.gsub(v, "_", " ")
                table.insert(pitcherSkills, v)
            end
        elseif nextCommand.code == "RECORD" then
            records = nextCommand.params
            table.remove(g_commands, 1)
        else
            break
        end
    end
    
    return batterSkills, pitcherSkills, records
end

-- 타자 등장
function processBat(command)
    g_currBatter = g_isUp and g_awayPlayers[command.params[1]] or g_homePlayers[command.params[1]]

    local playerId = tonumber(command.params[1])
    g_currBatterId = playerId

    g_preventForceRun = false

    -- 타자 스킬 및 기록, 투수 스킬 얻기
    local batterSkills, pitcherSkills, records = processSkills()
    
    -- 타자 타석에 위치
    live_call("showBatter", g_isUp, playerId, records)

    -- 타자 스킬 보여주기
    showBatterSkills(batterSkills)

    -- 투수 스킬 보여주기
    showPitcherSkills(pitcherSkills)
    
    showMent("BAT")
end

-- 타자 스킬 보여주기
function showBatterSkills(batterSkills)
    if #batterSkills == 0 then
        return
    end
    
    live_call("showBatterSkills", batterSkills)
end

-- 투수 스킬 보여주기. 기존과 동일하면 skip
function showPitcherSkills(pitcherSkills)
    if #pitcherSkills == 0 then
        return
    end

    live_call("showPitcherSkills", pitcherSkills)
end

-- 볼카운트
function processBallCount(command)
    g_ballCnt = tonumber(command.params[1])
    g_strikeCnt = tonumber(command.params[2])
    g_outCnt = tonumber(command.params[3])
    
    live_call("showScoreboard", g_inning, g_isUp, g_homeScore, g_awayScore, g_ballCnt, g_strikeCnt, g_outCnt)
    
    if g_state == STATE_BAT_AGAIN then
        g_state = STATE_INIT
        
        showMent("BAT_AGAIN")
        
        return true
    end
    
    return false
end

-- Hit 처리
function processHit(command)
    g_state = STATE_HIT

    local pos = command.params[1]
    local type = command.params[3]

    -- Hit 처리시 사용될 context
    g_hitContext = {code = command.params[2], pos = pos, type = type, phase = 0}

    if command.params[2] == "IF_FLY" then
        -- 내야 플라이
        processHit_IF_FLY(pos, type, g_hitContext)
    elseif command.params[2] == "IF_FFLY" then
        -- 내야 파울플라이
        processHit_IF_FFLY(pos, type, g_hitContext)
    elseif command.params[2] == "IF_GRO" then
        -- 내야 땅볼
        processHit_IF_GRO(pos, type, g_hitContext)
    elseif command.params[2] == "IF_LD" then
        -- 내야 라인드라이브
        processHit_IF_LD(pos, type, g_hitContext)
    elseif command.params[2] == "IF_HIT" then
        -- 내야 안타
        processHit_IF_HIT(pos, type, g_hitContext)
    elseif command.params[2] == "IF_SBUNT" then
        -- 내야 기습번트
        processHit_IF_SBUNT(pos, type, g_hitContext)
    elseif command.params[2] == "OF_HIT1" then
        -- 외야 안타
        processHit_OF_HIT1(pos, type, g_hitContext)
    elseif command.params[2] == "OF_HIT2" or command.params[2] == "OF_HIT3" then
        -- 2루타, 3루타
        processHit_OF_HIT2(pos, type, g_hitContext)
    elseif command.params[2] == "OF_FLY" then
        -- 외야 플라이
        processHit_OF_FLY(pos, type, g_hitContext)
    elseif command.params[2] == "OF_FFLY" then
        -- 외야 파울플라이
        processHit_OF_FFLY(pos, type, g_hitContext)
    elseif command.params[2] == "OF_HR" or command.params[2] == "OF_ISPHR" then
        -- 홈런
        processHit_OF_HR(pos, type, g_hitContext)
    end
end

function processHit_IF_FLY(pos, type, ctx)
    local nextCommand = g_commands[1]
    if nextCommand.code == "ERR" then
        -- 에러 상황
        table.remove(g_commands, 1)

        ctx.isError = true
        
        showStillCut("GF_" .. pos .. "E_1_s")

        showMent("IF_FLY_ERR", {defName = getDefPlayer(pos).name})
    else
        local stillCutName = "GF_" .. pos .. "_1_s"
        
        showStillCut(stillCutName)
        
        showMent("IF_FLY", {pos = g_posName[pos]})
    end
end

function processHit_IF_FFLY(pos, type, ctx)
    showStillCut("GFF_" .. pos .. "_1_s")
    showMent("IF_FFLY_" .. pos)
end

function processHit_IF_GRO(pos, type, ctx)
    local nextCommand = g_commands[1]
    if nextCommand.code == "ERR" then
        -- 에러 상황
        table.remove(g_commands, 1)
        
        ctx.isError = true
        
        showStillCut("GB_" .. pos .. "E_1_s")
        
        showMent("IF_GRO_ERR", {defName = getDefPlayer(pos).name})
    else
        local mentCode = "IF_GRO"
        local stillCutName = "GB_" .. pos

        if pos == "B1" or pos == "P" or pos == "C" then
            -- 1루, 투수, 포수는 easy/hard 구분 없음
            stillCutName = stillCutName .. "_1_s"
        else
            -- 2루, 3루, 유격수는 easy/hard에 따라 스틸컷 다름
            if type == "E" then
                stillCutName = stillCutName .. "_1_s"
            else
                mentCode = "IF_GRO_" .. pos
                stillCutName = stillCutName .. "_2_s"
            end
        end
        
        showStillCut(stillCutName)
        
        showMent(mentCode, {pos = g_posName[pos]})
    end
end

function processHit_IF_LD(pos, type, ctx)
    local nextCommand = g_commands[1]
    if nextCommand.code == "ERR" then
        -- 에러 상황
        table.remove(g_commands, 1)
        
        ctx.isError = true
        
        showStillCut("GF_" .. pos .. "E_1_s")
        
        showMent("IF_LD_ERR", {defName = getDefPlayer(pos).name})
    else
        if pos == "P" then
            showStillCut("GF_P_1_s")
        else
            showStillCut("LD_" .. pos .. "_1_s")
        end
        
        showMent("IF_LD", {pos = g_posName[pos]})
    end
end

function processHit_IF_HIT(pos, type, ctx)
    local mentCode = "IF_HIT_" .. pos
    local stillCutName = "GB_" .. pos .. "_2_s"
    
    showStillCut(stillCutName)
    
    showMent(mentCode)
end

function processHit_IF_SBUNT(pos, type, ctx)
    showStillCut("GB_P_1_s")
    
    showMent("IF_SBUNT_1")
end

function processHit_OF_HIT1(pos, type, ctx)
    local stillCutName = "HIT_" .. pos
    if type == "E" then
        stillCutName = stillCutName .. "_2_s"
    else
        stillCutName = stillCutName .. "_1_s"
    end
    
    showStillCut(stillCutName)
    
    local mentCode = "OF_HIT1_" .. pos .. "_" .. type
    showMent(mentCode)
end

function processHit_OF_HIT2(pos, type, ctx)
    local stillCutName = "2B_" .. pos
    local mentCode = "OF_HIT2_" .. pos

    if pos == "LF" or pos == "RF" then
        local idx = math.random(1, 3)
        stillCutName = stillCutName .. "_" .. idx .. "_s"
        mentCode = mentCode .. "_" .. idx
    else
        stillCutName = stillCutName .. "_1_s"
    end
    
    showStillCut(stillCutName)
    showMent(mentCode)
end

function processHit_OF_FLY(pos, type, ctx)
    if ctx.isHard then
        -- 깊은 외야 플라이일 경우
        showStillCut("OF_" .. pos .. "_1_s")
        
        if isMyAtk() then
            showMent("OF_FLY_5")
        else
            showMent("OF_FLY_6")
        end
    else
        local nextCommand = g_commands[1]
        if nextCommand.code == "ERR" then
            -- 에러 상황
            table.remove(g_commands, 1)

            ctx.isError = true
            
            showStillCut("OF_" .. pos .. "E_s")
            
            showMent("OF_FLY_ERR", {defName = getDefPlayer(pos).name})
        else
            -- 펜스에서 잡히는 타구인지 결정
            local isHard = type == "H" and math.random() < 0.05

            if not isHard then
                showStillCut("OF_" .. pos .. "_1_s")
                
                showMent("OF_FLY_1", {pos = g_posName[pos]})
            else
                ctx.isHard = true

                -- 타구 날라가는 스틸컷 보여주기
                if pos == "LF" then
                    showStillCut("BIGBALL_1_s")
                elseif pos == "CF" then
                    showStillCut("BIGBALL_2_s")
                else
                    showStillCut("BIGBALL_3_s")
                end

                -- 타자 이펙트 등장
                live_call("showBatterEffect")

                showMent("OF_FLY_4", {fence = g_fenceName[pos]})
            end
        end
    end
end

function processHit_OF_FFLY(pos, type, ctx)
    showStillCut("OF_" .. pos .. "_2_s")
    showMent("OF_FFLY_1", {pos = g_posName[pos]})
end

function processHit_OF_HR(pos, type, ctx)
    if ctx.isFly then
        if pos == "LF" then
            showStillCut("HR_1_s")
        elseif pos == "CF" then
            showStillCut("HR_2_s")
        else
            showStillCut("HR_3_s")
        end

        local runnerCnt = getRunnerCnt()
        if runnerCnt == 3 then
            showMent("OF_HR_2")
        elseif runnerCnt == 2 then
            showMent("OF_HR_3")
        elseif runnerCnt == 1 then
            showMent("OF_HR_4")
        else
            showMent("OF_HR_5")
        end
    else
        -- 타구 날라가는 스틸컷 보여주기
        if pos == "LF" then
            showStillCut("BIGBALL_1_s")
        elseif pos == "CF" then
            showStillCut("BIGBALL_2_s")
        else
            showStillCut("BIGBALL_3_s")
        end

        -- 타자 이펙트 등장
        live_call("showBatterEffect")
        
        showMent("OF_HR_1", {fence = g_fenceName[pos]})
    
        ctx.isFly = true
    end
end

function getRunnerCnt()
    local runnerCnt = 0

    for i = 1, 3 do
        if g_bases[i].runner then
            runnerCnt = runnerCnt + 1
        end
    end
    
    return runnerCnt
end

function postProcess()
    local result = {}

    result.runnerCnt = getRunnerCnt()
    result.outCnt = g_outCnt

    while #g_commands > 0 do
        local c = g_commands[1]
        if c.code == "HT" then
            g_history = c.params[1]

            g_state = STATE_INIT
            removeCommands()
            break
        elseif c.code == "C" or c.code == "PITCNG" then
            g_state = STATE_BAT_AGAIN
            break
        elseif c.code == "BS" or c.code == "BT" or c.code == "PS" or c.code == "PT" then
            -- 도루, 견제에의해서 스킬이 변경될 때 여기서 끊어줌
            g_state = STATE_BAT_AGAIN
            break
        elseif c.code == "OUT" then
            -- 아웃카운트 알림
            g_outCnt = tonumber(c.params[1])
            live_call("setOutCnt", g_outCnt)
        elseif c.code == "BRUN" or c.code == "RUN" then
            -- 타자 or 주자 움직임
            local fromBase
            local toBase
            local isSafeVisible
            
            if c.code == "BRUN" then
                fromBase = 0
                toBase = tonumber(c.params[1])
                isSafeVisible = c.params[2] == "T"
            else
                fromBase = tonumber(c.params[1])
                toBase = tonumber(c.params[2])
                isSafeVisible = c.params[3] == "T"
            end
            
            if toBase == 4 then
                -- 만일 홈에서 기다리는 상황이라면 일단은 아웃 처리하고 1루에 flag 세팅
                if c.code == "RUN" and c.params[3] == "W" then
                    moveRunnerToHome(fromBase, true, nil, false)
                    g_decideRun = true
                else
                    moveRunnerToHome(fromBase, false, nil, isSafeVisible)
                end
            else
                moveRunner(fromBase, toBase, false, nil, isSafeVisible)
            end
        elseif c.code == "RUNOUT" then
            -- 주자 아웃
            local fromBase = tonumber(c.params[1])
            local toBase = tonumber(c.params[2])
            local outType = c.params[3]
            
            if fromBase == 0 and outType == "A" then
                -- 타자 움직임 없이 아웃될 경우 (삼진, 플라이등)
                live_call("moveRunner", 0, 0)
            else
                if toBase == 4 then
                    moveRunnerToHome(fromBase, true, outType, false)
                else
                    moveRunner(fromBase, toBase, true, outType, false)
                end
            end
            
            -- 병살 시도 여부
            if outType == "D" then
                result.isDoublePlay = true

                g_preventForceRun = true
            end
        elseif c.code == "H" or c.code == "B" or c.code == "E" then
            -- 전광판 업데이트
            updateScoreboard(c)

	    	if c.code == "H" and c.params[3] == "T" then
				g_hitContext.isWalkOff = "T"
            end
        elseif c.code == "P" then
            -- 병살 시도일 경우에만 stop
            if result.isDoublePlay then
                table.remove(g_commands, 1)
                break
            end
        end
        
        table.remove(g_commands, 1)
    end
    
    -- 루상에 주자가 겹쳤을 때 주자 움직임 처리

    -- 3아웃에 타자가 삼진에 의해서 아웃되는 상황이 아닐 경우 무조건 뛴다
    local forceRun = not g_preventForceRun and g_outCnt == 3 and not g_bases[1].finished
    for i = 1, 3 do
        if (not g_bases[i].finished and g_bases[i].runner) or (forceRun and g_bases[i].runner) then
            if i < 3 then
                moveRunner(i, i + 1, false, nil, false)
            else
                moveRunnerToHome(i, true, nil, false)
            end
        end
    end

    -- TODO 홈에서 먼저 온 주자에 태그가 되는 순서 변경
    
    return result
end

function postProcessHit()
    if g_hitContext.isFly then
        -- 홈런일 경우 타구가 두단계에 걸쳐서 보여줌
        processHit_OF_HR(g_hitContext.pos, g_hitContext.type, g_hitContext)

        g_hitContext.isFly = nil
        return
    elseif g_hitContext.isHard then
        -- 외야 깊은 플라이일 경우 두단계에 걸쳐서 보여줌
        processHit_OF_FLY(g_hitContext.pos, g_hitContext.type, g_hitContext)
        
        g_hitContext.isHard = nil
        return
    end

	g_hitContext.isWalkOff = "F"

    -- 주자 여부
    g_hitContext.existRunner = {}
    g_hitContext.existRunner[1] = g_bases[1].runner
    g_hitContext.existRunner[2] = g_bases[2].runner
    g_hitContext.existRunner[3] = g_bases[3].runner

    g_hitContext.runnableCnt = 0
    if g_hitContext.existRunner[2] and g_hitContext.existRunner[3] then
        g_hitContext.runnableCnt = 2
    elseif g_hitContext.existRunner[2] or g_hitContext.existRunner[3] then
        g_hitContext.runnableCnt = 1
    end
    
    local result = postProcess()

    g_hitContext.runnerCnt = result.runnerCnt
    g_hitContext.outCnt = result.outCnt
    g_hitContext.phase = g_hitContext.phase + 1

    if g_hitContext.code == "IF_FLY" then
        -- 내야 플라이
        postProcessHit_IF_FLY(g_hitContext)
    elseif g_hitContext.code == "IF_FFLY" then
        -- 내야 파울플라이
        postProcessHit_IF_FFLY(g_hitContext)
    elseif g_hitContext.code == "IF_GRO" then
        -- 내야 땅볼
        
        -- 2아웃 내야땅볼에서 타자가 움직이지 않을 경우 강제 움직임
        if not g_hitContext.isError and g_hitContext.outCnt == 2 and g_bases[1].finished then
            moveRunner(0, 1, false, nil, false)
        end

        postProcessHit_IF_GRO(g_hitContext)
    elseif g_hitContext.code == "IF_LD" then
        -- 내야 라인드라이브
        postProcessHit_IF_LD(g_hitContext)
    elseif g_hitContext.code == "IF_HIT" then
        -- 내야 안타
        postProcessHit_IF_HIT(g_hitContext)
    elseif g_hitContext.code == "IF_SBUNT" then
        -- 내야 기습번트
        postProcessHit_IF_SBUNT(g_hitContext)
    elseif g_hitContext.code == "OF_HIT1" then
        -- 외야 안타
        postProcessHit_OF_HIT1(g_hitContext)
    elseif g_hitContext.code == "OF_HIT2" then
        -- 외야 2루타
        postProcessHit_OF_HIT2(g_hitContext)
    elseif g_hitContext.code == "OF_HIT3" then
        -- 외야 3루타
        postProcessHit_OF_HIT3(g_hitContext)
    elseif g_hitContext.code == "OF_FLY" then
        -- 외야 플라이
        postProcessHit_OF_FLY(g_hitContext)
    elseif g_hitContext.code == "OF_FFLY" then
        -- 외야 파울플라이
        postProcessHit_OF_FFLY(g_hitContext)
    elseif g_hitContext.code == "OF_HR" or g_hitContext.code == "OF_ISPHR" then
        -- 홈런, 인사이드 파크 홈런
        postProcessHit_OF_HR(g_hitContext)
    end
end

-- 내야플라이 후처리
function postProcessHit_IF_FLY(ctx)
    if ctx.isError then
        -- 홈 득점 여부
        if g_bases[4].queue and #g_bases[4].queue > 0 then
            addMentQueue(addPostfix("IF_FLY_CMT_1"))
        else
            addMentQueue(addPostfix("IF_FLY_CMT_2"))
        end
    else
        addMentQueue(addPostfix("IF_FLY_CMT_3"))
    end
end

-- 내야파울플라이 후처리
function postProcessHit_IF_FFLY(ctx)
    addMentQueue(addPostfix("IF_FFLY_CMT"))
end

-- 내야 땅볼 후처리
function postProcessHit_IF_GRO(ctx)
    if ctx.isError then
        -- 에러 상황
        if g_bases[4].queue and #g_bases[4].queue > 0 then
            addMentQueue("IF_GRO_CMT_5")
        else
            addMentQueue("IF_GRO_CMT_6")
        end
    else
        if ctx.phase == 1 then
            if ctx.runnerCnt > 0 then
                -- 송구 여부 얻기
                local throwBase = getThrowBase()
                if throwBase then
                    if throwBase.base == 4 then
                        ctx.isHomeThrow = true
                        
                        if throwBase.isOut then
                            if throwBase.outType == "T" then
                                addMentQueue("IF_GRO_1")
                            else
                                -- 병살 상황
                                addMentQueue("IF_GRO_2")
                            end
                        else
                            addMentQueue("IF_GRO_3")
                            addMentQueue("IF_GRO_CMT_4")
                        end
                    elseif throwBase.base == 3 then
                        addMentQueue("IF_GRO_6")
                    elseif throwBase.base == 2 then
                        addMentQueue("IF_GRO_7")
                    else
                        -- 1루 아웃
                        if throwBase.outType == "F" then
                            addMentQueue("IF_GRO_8")
                        else
                            addMentQueue("IF_GRO_9")
                        end
                        
                        -- 홈으로 들어왔는가?
                        if isHomeSafe() then
                            addMentQueue("IF_GRO_CMT_3")
                        end
                    end
                end
                
                -- 홈에 들어왔는지 여부
                ctx.isEnterHome = not g_bases[4].finished
            else
                -- 주자 없을 경우
                if g_bases[1].outType == "S" then
                    addMentQueue("IF_GRO_10")
                else
                    addMentQueue("IF_GRO_11")
                end
            end
        else
            -- 병살 상황
            if g_bases[1].isOut then
                addMentQueue("IF_GRO_4")
                
                if ctx.outCnt == 2 then
                    addMentQueue("IF_GRO_CMT_1")
                else
                    addMentQueue("IF_GRO_CMT_2")
                end
            else
                addMentQueue("IF_GRO_5")
                
                if not ctx.isHomeThrow and ctx.isEnterHome then
                    addMentQueue("IF_GRO_CMT_3")
                end
            end
        end
    end
end

-- 내야 라인드라이브 후처리
function postProcessHit_IF_LD(ctx)
    if ctx.isError then
        -- 에러 상황
        if g_bases[4].queue and #g_bases[4].queue > 0 then
            addMentQueue("IF_LD_CMT_1")
        else
            addMentQueue("IF_LD_CMT_2")
        end
    else
        addMentQueue("IF_LD_CMT_3")
    end
end

-- 내야안타 후처리
function postProcessHit_IF_HIT(ctx)
    addMentQueue("IF_HIT_1")
    
    if ctx.isWalkOff == "T" then
        addMentQueue("HIT_OVER")
    elseif ctx.existRunner[3] then
        addMentQueue(addPostfix("IF_HIT_CMT_1"))
    else
        addMentQueue(addPostfix("IF_HIT_CMT_2"))
    end
end

-- 내야 기습번트 후처리
function postProcessHit_IF_SBUNT(ctx)
    addMentQueue("IF_SBUNT_2")
    
    if ctx.isWalkOff == "T" then
        addMentQueue("HIT_OVER")
    elseif ctx.existRunner[3] then
        addMentQueue(addPostfix("IF_SBUNT_CMT_1"))
    else
        addMentQueue(addPostfix("IF_SBUNT_CMT_2"))
    end
end

-- 외야 안타 후처리
function postProcessHit_OF_HIT1(ctx)
    if ctx.isWalkOff == "T" then
        addMentQueue("HIT_OVER")
    elseif ctx.existRunner[3] then
        -- 선행 주자 3루
        if ctx.existRunner[2] then
            if #g_bases[4].queue > 1 then
                -- 2루에서 홈
                if not g_bases[4].queue[2].isOut then
                    addMentQueue("OF_HIT1_1")
                    addMentQueue(addPostfix("OF_HIT1_CMT_1"))
                else
                    addMentQueue("OF_HIT1_2")
                    addMentQueue(addPostfix("OF_HIT1_CMT_2"))
                end
            else
                -- 2루에서 3루
                addMentQueue(addPostfix("OF_HIT1_CMT_3"))
            end
        elseif ctx.existRunner[1] then
            if not g_bases[3].finished then
                -- 1루에서 3루
                if not g_bases[3].isOut then
                    addMentQueue("OF_HIT1_3")
                    addMentQueue(addPostfix("OF_HIT1_CMT_4"))
                else
                    addMentQueue("OF_HIT1_4")
                    addMentQueue(addPostfix("OF_HIT1_CMT_5"))
                end
            else
                -- 1루에서 2루
                addMentQueue(addPostfix("OF_HIT1_CMT_6"))
            end
        else
            -- 다음 주자 없음
            addMentQueue(addPostfix("OF_HIT1_CMT_7"))
        end
    elseif ctx.existRunner[2] then
        if g_bases[4].queue and #g_bases[4].queue > 0 then
            -- 2루주자 홈으로 들어오는가
            if not g_bases[4].queue[1].isOut then
                -- 세이프
                addMentQueue("OF_HIT1_5")
                addMentQueue(addPostfix("OF_HIT1_CMT_8"))
            else
                -- 아웃
                addMentQueue("OF_HIT1_6")
                addMentQueue(addPostfix("OF_HIT1_CMT_9"))
            end
        else
            addMentQueue(addPostfix("OF_HIT1_CMT_10"))
        end
    elseif ctx.existRunner[1] then
        if not g_bases[3].finished then
            -- 1루 주자 3루로 감
            if not g_bases[3].isOut then
                -- 세이프
                addMentQueue("OF_HIT1_7")
                addMentQueue(addPostfix("OF_HIT1_CMT_11"))
            else
                -- 아웃
                addMentQueue("OF_HIT1_8")
                addMentQueue(addPostfix("OF_HIT1_CMT_12"))
            end
        else
            addMentQueue(addPostfix("OF_HIT1_CMT_13"))
        end
    else
        -- 선행 주자 없을 경우
        addMentQueue(addPostfix("OF_HIT1_CMT_14"))
    end
end

-- 2루타 후처리
function postProcessHit_OF_HIT2(ctx)
    if ctx.isWalkOff == "T" then
        addMentQueue("HIT_OVER")
    elseif ctx.existRunner[2] or ctx.existRunner[3] then
        -- 득점권 주자 있음
        if ctx.existRunner[1] then
            -- 1루 주자 있음
            local last = g_bases[4].queue[#g_bases[4].queue]
            if last.fromBase == 1 then
                if not last.isOut then
                    addMentQueue("OF_HIT2_1", {run = ctx.runnableCnt + 1})
                    addMentQueue(addPostfix("OF_HIT2_CMT_1"))
                else
                    addMentQueue("OF_HIT2_2", {run = ctx.runnableCnt})
                    addMentQueue(addPostfix("OF_HIT2_CMT_2"))
                end
            else
                addMentQueue("OF_HIT2_3", {run = ctx.runnableCnt})
                addMentQueue(addPostfix("OF_HIT2_CMT_3"))
            end
        else
            addMentQueue("OF_HIT2_4", {run = ctx.runnableCnt})
            addMentQueue(addPostfix("OF_HIT2_CMT_4"))
        end
    else
        if ctx.existRunner[1] then
            -- 1루 주자 있음
            if g_bases[4].queue and #g_bases[4].queue > 0 then
                if not g_bases[4].queue[1].isOut then
                    -- 1루 주자 홈에서 세이프
                    addMentQueue("OF_HIT2_5")
                    addMentQueue(addPostfix("OF_HIT2_CMT_5"))
                else
                    -- 1루 주자 홈에서 아웃
                    addMentQueue("OF_HIT2_6")
                    addMentQueue(addPostfix("OF_HIT2_CMT_6"))
                end
            else
                addMentQueue("OF_HIT2_7")
                addMentQueue(addPostfix("OF_HIT2_CMT_7"))
            end
        else
            -- 주자 없음
            addMentQueue("OF_HIT2_8")
            addMentQueue(addPostfix("OF_HIT2_CMT_8"))
        end
    end
end

-- 3루타 후처리
function postProcessHit_OF_HIT3(ctx)
    if ctx.isWalkOff == "T" then
        addMentQueue("HIT_OVER")
    elseif ctx.runnerCnt > 0 then
        -- 주자 있음
        addMentQueue("OF_HIT3_1", {run = ctx.runnerCnt})
        addMentQueue(addPostfix("OF_HIT3_CMT_1"))
    else
        -- 주자 없음
        addMentQueue("OF_HIT3_2")
        addMentQueue(addPostfix("OF_HIT3_CMT_2"))
    end
end

-- 외야 플라이 후처리
function postProcessHit_OF_FLY(ctx)
    if ctx.isError then
        -- 에러 상황
        if g_bases[4].queue and #g_bases[4].queue > 0 then
            -- 3루주자 홈으로 들어온 상황
            addMentQueue(addPostfix("OF_FLY_CMT_1"))
        else
            addMentQueue(addPostfix("OF_FLY_CMT_2"))
        end
    else
        if ctx.outCnt == 2 then
            -- 2아웃일 경우
            addMentQueue(addPostfix("OF_FLY_CMT_3"))
        else
            if ctx.runnerCnt > 0 then
                -- 선행주자 있을 경우
                local runner = getMoveRunner()
                if runner then
                    if not runner.isOut then
                        -- 세이프
                        addMentQueue("OF_FLY_2", {fromBase = g_baseName[runner.fromBase], toBase = g_baseName[runner.toBase]})
                        
                        if runner.fromBase == 3 then
                            addMentQueue(addPostfix("OF_FLY_CMT_4"))
                        else
                            addMentQueue(addPostfix("OF_FLY_CMT_5"))
                        end
                    else
                        -- 아웃
                        addMentQueue("OF_FLY_3", {fromBase = g_baseName[runner.fromBase], toBase = g_baseName[runner.toBase]})

                        addMentQueue(addPostfix("OF_FLY_CMT_6"))
                    end
                else
                    -- 주자 움직임 없음
                    addMentQueue("OF_FLY_CMT_7")
                end
            else
                -- 선행주자 없을 경우
                addMentQueue(addPostfix("OF_FLY_CMT_8"))
            end
        end
    end
end

-- 외야 파울플라이 후처리
function postProcessHit_OF_FFLY(ctx)
    if ctx.outCnt == 2 then
        -- 2아웃일 경우
        addMentQueue(addPostfix("OF_FFLY_CMT_1"))
    else
        if ctx.runnerCnt > 0 then
            -- 선행주자 있을 경우
            local runner = getMoveRunner()
            if runner then
                if not runner.isOut then
                    -- 세이프
                    addMentQueue("OF_FFLY_2", {fromBase = g_baseName[runner.fromBase], toBase = g_baseName[runner.toBase]})
                    
                    if runner.fromBase == 3 then
                        addMentQueue(addPostfix("OF_FFLY_CMT_2"))
                    else
                        addMentQueue(addPostfix("OF_FFLY_CMT_3"))
                    end
                else
                    -- 아웃
                    addMentQueue("OF_FFLY_3", {fromBase = g_baseName[runner.fromBase], toBase = g_baseName[runner.toBase]})
                    
                    addMentQueue(addPostfix("OF_FFLY_CMT_4"))
                end
            else
                -- 주자 움직임 없음
                addMentQueue("OF_FFLY_CMT_5")
            end
        else
            addMentQueue(addPostfix("OF_FFLY_CMT_6"))
        end
    end
end

-- 홈런 처리
function postProcessHit_OF_HR(ctx)
    if ctx.isWalkOff == "T" then
        addMentQueue("HR_OVER")
    elseif ctx.runnerCnt == 3 then
        addMentQueue(addPostfix("OF_HR_CMT_1"))
    elseif ctx.runnerCnt == 2 then
        addMentQueue(addPostfix("OF_HR_CMT_2"))
    elseif ctx.runnerCnt == 1 then
        addMentQueue(addPostfix("OF_HR_CMT_3"))
    else
        addMentQueue(addPostfix("OF_HR_CMT_4"))
    end
end

-- 삼진 처리
function processSo(command)
    g_state = STATE_SO

    if isMyAtk() then
        showMent("SO_M")
    else
        showMent("SO_O")
    end
    
    showStillCut("MENT_SO_NEW")
end

function postProcessSo(command)
    -- 주자, 아웃 카운트 처리 전에 멘트 결정
    if isMyAtk() then
        if isRunnableBase() then
            if g_outCnt == 2 then
                addMentQueue("SO_CMT_1")
            else
                addMentQueue("SO_CMT_2")
            end
        else
            if g_outCnt == 2 then
                addMentQueue("SO_CMT_3")
            else
                addMentQueue("SO_CMT_4")
            end
        end
    else 
        if getRunnerCnt() > 0 and g_outCnt == 2 then
            if isRunnableBase() then
                addMentQueue("SO_CMT_5")
            else
                addMentQueue("SO_CMT_6")
            end
        end
    end
    
    postProcess()
end

-- 볼넷 처리
function processBb(command)
    g_state = STATE_BB
    
    local runnerCnt = getRunnerCnt()

    if runnerCnt == 3 then
        if isMyAtk() then
            showMent("BB_1_M")
        else
            showMent("BB_1_O")
        end
    else
        if isMyAtk() then
            showMent("BB_2_M")
        else
            showMent("BB_2_O")
        end
    end
    
    showStillCut("MENT_BB_NEW")
end

function postProcessBb(command)
   local result = postProcess()
   
    if result.runnerCnt == 3 then
        addMentQueue(addPostfix("BB_CMT_1"))
    elseif result.runnerCnt == 2 then
        addMentQueue(addPostfix("BB_CMT_2"))
    elseif result.runnerCnt == 1 then
        addMentQueue(addPostfix("BB_CMT_3"))
    else
        addMentQueue(addPostfix("BB_CMT_4"))
    end
end

-- 사구 
function processHbp(command)
    g_state = STATE_HBP
    
    local runnerCnt = getRunnerCnt()

    if runnerCnt == 3 then
        -- 만루 상황
        if isMyAtk() then
            showMent("HBP_1")
        else
            showMent("HBP_2")
        end
    else
        -- 만루가 아닐 경우
        if isMyAtk() then
            showMent("HBP_3")
        else
            showMent("HBP_4")
        end
    end
end

function postProcessHbp(command)
    local result = postProcess()
    
    if result.runnerCnt == 3 then
        addMentQueue(addPostfix("HBP_CMT_1"))
    elseif result.runnerCnt == 2 then
        addMentQueue(addPostfix("HBP_CMT_2"))
    elseif result.runnerCnt == 1 then
        addMentQueue(addPostfix("HBP_CMT_3"))
    else
        addMentQueue(addPostfix("HBP_CMT_4"))
    end
end

-- 낫아웃
function processNotout(command)
    g_state = STATE_NOTOUT

    showStillCut("NOT_OUT_s")
    
    showMent("NOTOUT_1")
end

-- 낫아웃 후처리
function postProcessNotout(command)
    local result = postProcess()
    
    -- NOTOUT- 에서 타자주자 움직임이 없기 때문에 여기서 강제로 넣어줌
    if g_bases[1].finished then
        moveRunner(0, 1, true, "F", false)
        
        -- 루상에 주자가 겹쳤을 때 주자 움직임 처리
        for i = 1, 3 do
            if not g_bases[i].finished and g_bases[i].runner then
                if i < 3 then
                    moveRunner(i, i + 1, false, nil, false)
                else
                    moveRunnerToHome(i, true, nil, false)
                end
            end
        end
    end
    
    if not g_bases[1].isOut then
        addMentQueue("NOTOUT_2")
        
        if result.runnerCnt == 3 then
            addMentQueue("NOTOUT_CMT_1")
        else
            addMentQueue("NOTOUT_CMT_2")
        end
    else
        addMentQueue("NOTOUT_3")
        
        addMentQueue("NOTOUT_CMT_3")
    end
end

-- 희생번트
function processSbnt(command)
    showStillCut("BUNT_s")
    showMent("SBNT_1")
    
    g_state = STATE_SBNT
end

-- 희생번트 결과
function postProcessSbnt()
    local result = postProcess()

    local outRunner = getOutRunner()
    if outRunner then
        -- 희생번트 실패
        addMentQueue("SBNT_2", {fromBase = g_baseName[outRunner.fromBase], toBase = g_baseName[outRunner.toBase]})
        
        addMentQueue(addPostfix("SBNT_CMT_1"))
    else
        -- 희생번트 성공
        addMentQueue("SBNT_3")

        addMentQueue(addPostfix("SBNT_CMT_2"))
    end
end

-- 도루 시도
function processStl(command)
    g_state = STATE_STL
    
    showStillCut("SB_" .. command.params[1] .. "_s")
    showStillCut("SB_HOME_s")
    
    local fromBase = tonumber(command.params[1])
    local toBase = fromBase + 1
    
    g_stealer = g_isUp and g_awayPlayers[g_bases[fromBase].runner] or g_homePlayers[g_bases[fromBase].runner]

    showMent("STL_1", {fromBase = g_baseName[fromBase], toBase = g_baseName[toBase], name = g_stealer.name})
end

-- 도루 결과 처리
function postProcessStl()
    local result = postProcess()
    
  	-- 도루 실패로 3아웃이 될 경우 
  	if g_state == STATE_STL and g_outCnt == 3 then
    	g_state = STATE_INIT
    end
  
    local runner = getMoveRunner()
    if not runner.isOut then
        -- 도루 성공
        addMentQueue("STL_2", {fromBase = g_baseName[runner.fromBase], toBase = g_baseName[runner.toBase], name = g_stealer.name})
        
        addMentQueue(addPostfix("STL_CMT_1"))
    else
        -- 도루 실패
        addMentQueue("STL_3", {fromBase = g_baseName[runner.fromBase], toBase = g_baseName[runner.toBase], name = g_stealer.name})
        
        addMentQueue(addPostfix("STL_CMT_2"))
    end
    
    g_stealer = nil
end

-- 견제
function processPickoff(command)
    g_state = STATE_PICKOFF

    g_pickoff = {base = tonumber(command.params[1]), isOut = false}

    showStillCut("PITCHER_PICKOFF_B" .. command.params[1])

    if command.params[2] == "T" then
        g_pickoff.isOut = true

        showStillCut("PICKOFF_OUT_B" .. command.params[1])

    else
        showStillCut("PICKOFF_SAFE_B" .. command.params[1])
    end

    showMent("PK")
end

-- 견제 결과 처리
function postProcessPickoff()
    local result = postProcess()

    -- 견제로 3아웃이 될 경우
    if g_state == STATE_PICKOFF and g_outCnt == 3 then
        g_state = STATE_INIT
    end

    -- 주자 아웃 처리
    if g_pickoff.isOut then
        local base = g_pickoff.base
        g_bases[base].runner = nil

        -- 주자 삭제
        live_call("removeRunner", base)

        showMent(addPostfix("PK_CMT"))
    else
        showMent(addPostfix("PT_CMT"))
    end

    g_pickoff = nil
end

-- 와일드피치
function processPitwp(command)
    g_state = STATE_PITWP
    
    showStillCut("WILD_PITCH_s")
    
    if g_bases[3].runner then
        if isMyAtk() then
            showMent("PITWP_1")
        else
            showMent("PITWP_2")
        end
    else
        if isMyAtk() then
            showMent("PITWP_3")
        else
            showMent("PITWP_4")
        end
    end
end

-- 와일드피치 결과 처리
function postProcessPitwp()
    local result = postProcess()

    if not g_bases[4].finished then
        addMentQueue(addPostfix("PITWP_CMT_1"))
    else
        addMentQueue(addPostfix("PITWP_CMT_2"))
    end
    
    -- 해당 타석에서 끝내기 폭투일 경우 다시 타석에 들어서는 C코드가 없어서 상태가 변하지 않음. 따라서 이 상황에서는 초기 상태로 변경
    if g_state == STATE_PITWP and #g_commands == 0 then
        g_state = STATE_INIT
    end
end

-- 투수 교체
function processPitcng(command)
    if g_isUp then
        g_homePitcherId = command.params[1]
    else
        g_awayPitcherId = command.params[1]
    end

    -- 현재 투수 세팅
    g_currPitcher = g_isUp and g_homePlayers[g_homePitcherId] or g_awayPlayers[g_awayPitcherId]
    
    live_call("changePitcher", g_isUp, tonumber(command.params[1]))

    showMent("PITCNG")
    
    if not isMyAtk() then
        local cpId = g_isHome and g_homeCpId or g_awayCpId
        if cpId == command.params[1] then
            -- 마무리 투수
            addMentQueue("PITCNG_CMT_1")
        else
            local diff = 0
            if g_isHome then
                diff = g_homeScore - g_awayScore
            else
                diff = g_awayScore - g_homeScore
            end
            
            if diff >= 1 then
                addMentQueue("PITCNG_CMT_2")
            elseif diff == 0 then
                addMentQueue("PITCNG_CMT_3")
            else
                addMentQueue("PITCNG_CMT_4")
            end
        end
    end
  
  	--addMentQueue("BAT_AGAIN")
  	g_state = STATE_BAT_AGAIN
end

function processGameover()
    g_state = STATE_GAMEOVER
    
    local myScore = g_isHome and g_homeScore or g_awayScore
    local otherScore = g_isHome and g_awayScore or g_homeScore

    local params = {teamName = g_isHome and g_homeTeam.teamName or g_awayTeam.teamName, myScore = myScore, otherScore = otherScore}

    if myScore > otherScore then
        showMent("GAMEOVER_1", params)
        
        if myScore - otherScore >= 5 then
            addMentQueue("GAMEOVER_CMT_1")
        else
            addMentQueue("GAMEOVER_CMT_2")
        end
    elseif myScore < otherScore then
        showMent("GAMEOVER_2", params)
        
        if myScore - otherScore <= -5 then
            addMentQueue("GAMEOVER_CMT_3")
        else
            addMentQueue("GAMEOVER_CMT_4")
        end
    else
        showMent("GAMEOVER_3", params)
        
        addMentQueue("GAMEOVER_CMT_5")
    end
end

-- 우리팀 공격 여부
function isMyAtk()
    return (g_isUp and not g_isHome) or (not g_isUp and g_isHome)
end

-- 득점권 상황 여부
function isRunnableBase()
    return g_bases[2].runner or g_bases[3].runner
end

-- 만루 여부
function isFullBase()
    return g_base[1].runner and g_base[2].runner and g_base[3].runner
end

-- 선행 주자 움직임
function getMoveRunner()
    if not g_bases[4].finished then
        local obj = g_bases[4].queue[1]
        return {fromBase = obj.fromBase, toBase = 4, isOut = obj.isOut}
    elseif not g_bases[3].finished then
        return {fromBase = g_bases[3].fromBase, toBase = 3, isOut = g_bases[3].isOut}
    elseif not g_bases[2].finished then
        return {fromBase = g_bases[2].fromBase, toBase = 2, isOut = g_bases[2].isOut}
    end
    
    return nil
end

-- 아웃된 선행주자 얻기
function getOutRunner()
    if not g_bases[2].finished and g_bases[2].isOut then
        return {fromBase = g_bases[2].fromBase, toBase = 2}
    elseif not g_bases[3].finished and g_bases[3].isOut then
        return {fromBase = g_bases[3].fromBase, toBase = 3}
    elseif not g_bases[4].finished then
        local obj = g_bases[4].queue[1]
        if obj.isOut then
            return {fromBase = obj.fromBase, toBase = obj.toBase}
        end
    end
    
    return nil
end

-- 송구 정보
function getThrowBase()
    -- 홈으로 송구
    if not g_decideRun and not g_bases[4].finished and #g_bases[4].queue > 0 then
        for i, v in ipairs(g_bases[4].queue) do
      		-- 홈에서 isOut과 outType이 함께 존재하는지 체크 해야함
      		-- 선행주자가 겹쳐서 홈으로 보낼때 isOut을 false로 주면 세이프로 간주하기 때문에 isOut을 true, outType을 nil로 세팅함
            if (v.isOut and v.outType) or v.isSafeVisible then
                return {base = 4, isOut = v.isOut, outType = v.outType, isSafeVisible = v.isSafeVisible}
            end
        end
    end

    -- 각 베이스 체크
    for i = 1, 3 do
        if not g_bases[i].finished and (g_bases[i].isOut or g_bases[i].isSafeVisible) then
            return {base = i, isOut = g_bases[i].isOut, outType = g_bases[i].outType, isSafeVisible = g_bases[i].isSafeVisible}
        end
    end

    return nil
end

-- 3루주자 홈으로 들어왔는가?
function isHomeSafe()
    if g_bases[4].queue and #g_bases[4].queue > 0 then
        for i, v in ipairs(g_bases[4].queue) do
            if not v.isOut then
                return true
            end
        end
    end
    
    return false
end

function addPostfix(mentCode)
    if isMyAtk() then
        return mentCode .. "_M"
    else
        return mentCode .. "_O"
    end
end

local g_defaultMentOptions = {
    delay = 1.25
}

function wrapColor(str, color, mentcolor)
    return "@{B,22," .. string.sub(color, 2, -2) .. ",255}" .. str .. "@{B,22," .. string.sub(mentcolor, 2, -2) .. "}"
end

function showMent(mentCode, params, options)
    local ment

    -- 해당 멘트가 string 또는 array인지 확인해서 array이면 랜덤하게 선택
    if type(g_ments[mentCode]) == "table" then
        local index = math.random(#g_ments[mentCode])
        --print(string.format("@@@@@@@ randomMent : %d @@@@@@@", index))
        ment = g_ments[mentCode][index]
    else
        ment = g_ments[mentCode]
    end
    
    -- 멘트가 만약 빈문자열이면 false 반환
    if not ment or #ment == 0 then
        return false
    end

    -- 깜빡임인가 체크
    local isBlink = false
    if #ment >= #BLINK and string.sub(ment, -#BLINK) == BLINK then
        isBlink = true
        ment = string.sub(ment, 1, -#BLINK - 1)
    end

    -- 색상 변경이 되는 코드일 경우 처리
    local color = g_mentColors[mentCode]
    if color then
        local isHomeColor = true
        if color == "ME" then
            isHomeColor = g_isHome
        elseif color == "ATK" then
            isHomeColor = not g_isUp
        elseif color == "DEF" then
            isHomeColor = g_isUp
        end
        
        if isHomeColor then
            g_fgColor = g_homeTeam.fgColor
            g_bgColor = g_homeTeam.bgColor
            g_mentColor = string.sub(g_homeTeam.mentColor, 1, -2) .. ",255)"
        else
            g_fgColor = g_awayTeam.fgColor
            g_bgColor = g_awayTeam.bgColor
            g_mentColor = string.sub(g_awayTeam.mentColor, 1, -2) .. ",255)"
        end
    end

    -- 템플릿 처리
    if not params then
        params = {}
    end

    -- 타자, 투수 이름 넣어주기
    if not params.name and g_currBatter then
        params.name = g_currBatter.name
    end
    
    if not params.pitcherName and g_currPitcher then
        params.pitcherName = g_currPitcher.name
    end

    if params.name then
        params.name = wrapColor(params.name, g_fgColor, g_mentColor)
    end
    
    if params.pitcherName then
        params.pitcherName = wrapColor(params.pitcherName, g_fgColor, g_mentColor)
    end
    
    if params.defName then
        params.defName = wrapColor(params.defName, g_fgColor, g_mentColor)
    end

    ment = string.gsub(ment, "{(%a+)}", params)

    -- 옵션 처리
    local o = {}
    for k, v in pairs(g_defaultMentOptions) do
        if options and options[k] then
            o[k] = options[k]
        else
            if k == "delay" then
                -- 멘트에 설정된 딜레이가 있을 경우 적용
                if g_delays[mentCode] then
                    o.delay = g_delays[mentCode]
                else
                    o[k] = v
                end
            else
                o[k] = v
            end
        end
    end

    --print(ment)
    
    live_call("showMent", ment, o.delay, g_fgColor, g_bgColor, g_mentColor, isBlink);
    
    return true
end

-- 멘트 큐에 추가
function addMentQueue(mentCode, params, options)
    table.insert(g_mentQueue, {mentCode = mentCode, params = params, options = options})
end

-- 주자 움직임(홈은 제외)
function moveRunner(fromBase, toBase, isOut, outType, isSafeVisible)
    g_bases[toBase].finished = false
    g_bases[toBase].fromBase = fromBase
    
    -- 해당 베이스에서 주자 떠남
    if fromBase > 0 then
        g_bases[fromBase].runner = nil
    end

    if isOut then
        g_bases[toBase].isOut = true
        g_bases[toBase].outType = outType
    else
        g_bases[toBase].isSafeVisible = isSafeVisible
    end

    live_call("moveRunner", fromBase, toBase)
end

-- 주자 홈으로 움직임
function moveRunnerToHome(fromBase, isOut, outType, isSafeVisible)
    g_bases[4].finished = false
    
    -- 해당 베이스 주자 떠남
    if fromBase > 0 then
        g_bases[fromBase].runner = nil
    end
    
    if g_bases[4].queue == nil then
        g_bases[4].queue = {}
    end
    
    local obj = {fromBase = fromBase}
    if isOut then
        obj.isOut = true
        obj.outType = outType
    else
        obj.isSafeVisible = isSafeVisible
    end

    table.insert(g_bases[4].queue, obj)
    
    live_call("moveRunner", fromBase, 4)
end

-- 주자 움직임 종료시 처리
function setRunnerFinished(base, playerId)
    --print("@@@@@@@@@@ " .. base)

    if base > 0 and not g_bases[base].finished then
        if base == 4 then
            -- 홈에서의 처리
            if g_bases[4].queue and #g_bases[4].queue > 0 then
                local obj = table.remove(g_bases[4].queue, 1)
                
                if obj.isOut then
                    -- 홈 아웃
                    if obj.outType then
                        showStillCut("OUT_HOME_" .. obj.outType .. "_s")
                    end
                else
                    -- 홈 세이프
                    
                    -- 스코어 증가
                    if g_isUp then
                        g_awayScore = g_awayScore + 1
                    else
                        g_homeScore = g_homeScore + 1
                    end
                    
                    -- 마지막 득점인지 체크
                    local isLast = true
                    for i, v in ipairs(g_bases[4].queue) do
                        if not v.isOut then
                            isLast = false
                            break
                        end
                    end

                    live_call("addScore", g_inning, g_isUp, isLast)
                    
                    if obj.isSafeVisible then
                        showStillCut("SAFE_HOME_s")
                    end
                end
                
                live_call("removeRunner", 4)
            end
            
            -- 홈에서 상황이 종료되었느지 체크
            g_bases[4].finished = g_bases[4].queue == nil or #g_bases[4].queue == 0
        else
            -- 1,2,3루 처리
            g_bases[base].finished = true

            if g_bases[base].isOut then
                -- 아웃
                
                -- 아웃 스틸컷이 있을 경우 표시
                if g_bases[base].outType then
                    showStillCut("OUT_B" .. base .. "_" .. g_bases[base].outType .. "_s")
                    g_bases[base].outType = nil
                end

                -- 주자 삭제
                --live_call("removeRunner", base)
                g_bases[base].isRemove = true
                g_bases[base].isOut = nil
            else
                -- 세이프
                if g_bases[base].isSafeVisible then
                    showStillCut("SAFE_B" .. base .. "_s")
                end
                
                g_bases[base].runner = playerId                
            end
        end
    end

    -- 모든 베이스에서 처리가 끝났는지 체크
    for i = 1, 4 do
        if not g_bases[i].finished then
            return
        end
    end
    
    -- 홈 득점 인정 처리
    if g_state == STATE_INIT and g_decideRun then
        if g_outCnt < 3 then
            -- 스코어 증가
            if g_isUp then
                g_awayScore = g_awayScore + 1
            else
                g_homeScore = g_homeScore + 1
            end
            
            live_call("addScore", g_inning, g_isUp, true)
        end
        
        g_decideRun = false
    end
    
    -- 주자 삭제. 모든 주자의 움직임이 끝나고 처리해야지 그렇지 않을 경우 주자 움직임 순서에 따라서 삭제 오류가 생김
    for i = 1, 3 do
        if g_bases[i].isRemove then
            --print("remove runner : " .. i)
            live_call("removeRunner", i)
            g_bases[i].isRemove = nil
        end
    end

    -- 타석 결과 알림
    if g_history then
        live_call("addHistory", g_isUp, g_currBatterId, g_history)
        g_history = nil
    end

    proceed()
end

function updateScoreboard(command)
    live_call("updateScoreboard", g_inning, command.params[1] == "U", command.code, tonumber(command.params[2]))
end

-- 스틸컷 보여주기
function showStillCut(stillCutName)
    live_call("showStillCut", stillCutName)
end