-- NORMAL 0
-- SEASON 1
-- SPECIAL 2
-- MONSTER 3
-- ACE 4

local g_cost = {}
g_cost[1] = {min = 40, max = 49}
g_cost[2] = {min = 50, max = 54}
g_cost[3] = {min = 55, max = 59}
g_cost[4] = {min = 60, max = 64}
g_cost[5] = {min = 65, max = 69}
g_cost[6] = {min = 70, max = 74}
g_cost[7] = {min = 75, max = 79}
g_cost[8] = {min = 80, max = 106}


-- 트레이드 결과 확률 계산
function calculate(players, useAp, useCash)
    print("players : " .. #players)
    
    local ap = 0
    local cash = 0
    
    -- 공식에서는 ap, cash값을 고정시킴
    if useAp then
        ap = 3000
    end
    
    if useCash then
        cash = 10
    end
    
    local minOverall = 0
    local maxOverall = 0
    local specialProb = 0
    local monsterProb = 0
    local isMonsterAvailable = false

    if #players < 2 then
        return minOverall, maxOverall, specialProb, monsterProb, isMonsterAvailable
    end
    
    -- 가장 높은 코스트
    local ace = -1
    -- 평균 코스트
    local avg = 0.0
    -- 몬스터 카드 개수
    local mon = 0
    -- 스페셜 카드 개수
    local spn = 0
    -- 스페셜 카드 평균 코스트
    local spa = 0.0
    -- 스페셜 카드의 가장 높은 코스트
    local spt = 0
    -- 노멀 카드 평균 코스트
    local non = 0

    -- 코스트 1, 2, 3 개수
    local cost1Cnt = 0
    local cost2Cnt = 0
    local cost3Cnt = 0
    
    -- 시즌 선수 개수
    local seasonCnt = 0

        -- 에이스 선수 갯수
        local aceCnt = 0;
        
    for i, p in ipairs(players) do
        avg = avg + p.cost

        if p.cost > ace then
            ace = p.cost
        end
        
        if p.cost == 1 then
            cost1Cnt = cost1Cnt + 1
        elseif p.cost == 2 then
            cost2Cnt = cost2Cnt + 1
        elseif p.cost == 3 then
            cost3Cnt = cost3Cnt + 1
        end

        if p.playerType == MONSTER then
            mon = mon + 1
        elseif p.playerType == ACE then
            aceCnt = aceCnt + 1
            
            -- 에이스 카드도 노멀과 같이 처리
            non = non + 1
        elseif p.playerType == SPECIAL then
            spn = spn + 1
            spa = spa + p.cost
            
            if p.cost > spt then
                spt = p.cost
            end
        elseif p.playerType == SEASON then
            seasonCnt = seasonCnt + 1

            -- 시즌 카드도 노멀과 같이 처리
            non = non + 1
        else
            non = non + 1
        end
    end
    
    avg = avg / #players
    if spn > 0 then
        spa = spa / spn
    end

    print("ace : " .. ace)
    print("avg : " .. avg)
    print("mon : " .. mon)
    print("spn : " .. spn)
    print("spa : " .. spa)
    print("non : " .. non)
    print("ap : " .. ap)
    print("cash : " .. cash)
    
    -- 몬스터 확률
    if mon > 0 then
        if useCash then
            monsterProb = 0.37 + (spn * spa * 0.004) * (cash * 0.32)
        else
            monsterProb = 0.37 + (spn * spa * 0.004)
        end

    elseif avg >= 6 then
        monsterProb = (spt - 6) * 0.016 + (spa * 2) * (spn * 0.5) * 0.0035

        if monsterProb < 0 then
            monsterProb = 0
        end

        if useCash then
            monsterProb = monsterProb * cash * 0.15
        end
    end
    
    -- 베이스 코스트 결정
    local baseCost = math.min(round(ace - (ace - avg) * 0.5 + spn * 0.2), 8)
    
    -- 3코 이하 같은 코스트 선수를 4명 투입했을 경우 처리
    if cost1Cnt == 4 or cost2Cnt == 4 or cost3Cnt == 4 then
        baseCost = ace + 1
    end

    -- 스페셜 확률
    if baseCost >= 6 then
        specialProb = 0.05 + spa * 0.02 + (spa * spn * 4) * 0.005
        if useCash then
            specialProb = specialProb * cash * 0.2
        end
        
        specialProb = math.min(specialProb, 1)
    elseif baseCost >= 3 then
        specialProb = 0.1 + spa * 0.04 + (spa * spn) * 0.01
        if useCash then
            specialProb = specialProb * cash * 0.2
        end
        
        specialProb = math.min(specialProb, 1)
    end

    local upCost = 1
    if useAp then
        upCost = 2
    end

    local dnCost = 2

    -- 3코 이하 같은 코스트 선수를 4명 투입했을 경우 처리
    if cost1Cnt == 4 or cost2Cnt == 4 or cost3Cnt == 4 then
        if useAp then
            upCost = 1
        else
            upCost = 0
        end

        dnCost = 0
    end

    -- 시즌 선수 레시피 특별 처리
    if seasonCnt == 4 then
        baseCost = 1
        dnCost = 0
        upCost = 0
        
        specialProb = 0
        monsterProb = 0
    elseif aceCnt == 4 then
        -- 에이스 선수 레시피 특별 처리
        baseCost = 4
        dnCost = 0
        upCost = 4
        
        specialProb = 0
        monsterProb = 0
    end

    local minCost = math.max(baseCost - dnCost, 1)
    local maxCost = math.min(baseCost + upCost, 8)
    
    minOverall = g_cost[minCost].min
    maxOverall = g_cost[maxCost].max

    specialProb = math.max(specialProb, 0)
    monsterProb = math.max(monsterProb, 0)

    -- 몬스터 등장 확률이 0.xxx로 나올 경우, 1%미만이라고 표기하기 위해 반올림 처리 전 몬스터 등장 가능 여부를 체크한다
    if monsterProb > 0 then
        isMonsterAvailable = true
    end
    
    return minOverall, maxOverall, round(specialProb * 100), round(monsterProb * 100), isMonsterAvailable
end

function round(num) 
    if num >= 0 then return math.floor(num+.5)
    else return math.ceil(num-.5) end
end
